from typing import Tuple
import numpy as np
import pandas
from scipy import stats
from ._recorders import NumpyArrayStorageRecorder, NumpyArrayNormalisedStorageRecorder


class GaussianKDEStorageRecorder(NumpyArrayStorageRecorder):
    """A recorder that fits a KDE to a time-series of volume.

    This recorder is an extension to `NumpyArrayStorageRecorder` which, at the end of a simulation,
    uses kernel density estimation (KDE) to estimate the probability density function of the storage time-series.
    It returns the probability of being at or below a specified target volume in the `aggregated_value()` method
    (i.e. used for optimisation). The recorder flattens data from all scenarios before computing the KDE. Therefore,
    a single PDF is produced and returned via `.to_dataframe()`.

    The user can specify an optional resampling (e.g. to create annual minimum time-series) prior to fitting
    the KDE. By default the KDE is reflected at the proportional storage bounds (0.0 and 1.0) to represent the
    boundedness of the distribution. This can be disabled if required.

    Parameters
    ==========
    resample_freq : str or None
        If not None the resampling frequency used by prior to distribution fitting.
    resample_func : str or None
        If not None the resampling function used prior to distribution fitting.
    target_volume_pc : float
        The proportional target volume for which a probability of being at or lower is estimated.
    num_pdf : int
        Number of points in the PDF estimate. Defaults to 101.
    use_reflection : bool
        Whether to reflect the PDF at the upper and lower bounds (i.e. 0% and 100% volume) to account for
        the boundedness of the distribution. Defaults to true.
    """
    def __init__(self, *args, **kwargs):
        self.resample_freq = kwargs.pop('resample_freq', None)
        self.resample_func = kwargs.pop('resample_func', None)
        self.target_volume_pc = kwargs.pop('target_volume_pc')
        self.use_reflection = kwargs.pop('use_reflection', True)
        self.num_pdf = kwargs.pop('num_pdf', 101)

        super().__init__(*args, proportional=True, **kwargs)
        self._probability_of_target_volume = None
        self._pdf = None

    def reset(self):
        super().reset()
        self._probability_of_target_volume = None
        self._pdf = None

    def finish(self):
        super().finish()

        df = super().to_dataframe()
        # Apply resampling if defined
        if self.resample_func is not None and self.resample_freq is not None:
            df = df.resample(self.resample_freq).agg(self.resample_func)

        x = np.linspace(0.0, 1.0, self.num_pdf)

        # Compute the probability for the target volume
        p, pdf = self._estimate_pdf_and_target_probability(df.values.flatten(), x)

        self._probability_of_target_volume = p
        self._pdf = pandas.DataFrame(data=pdf, index=x)

    def values(self):
        """Return the estimated PDF values."""
        return self._pdf.values

    def to_dataframe(self):
        """ Return a `pandas.DataFrame` of the estimated PDF.
        """
        return self._pdf

    def aggregated_value(self):
        return self._probability_of_target_volume

    def _estimate_pdf_and_target_probability(self, values, x) -> Tuple[float, np.ndarray]:
        """Return a probability of being at below `self.target_volume_pc` and a estimate of the PDF

        This method can (if `self.use_reflection` is truthy) reflect the PDF at the lower and upper boundaries
        to stop the PDF leaking in to infeasible space.
        """
        # Fit a Gaussian KDE
        kernel = stats.gaussian_kde(values)
        p = kernel.integrate_box_1d(0.0, self.target_volume_pc)
        pdf = kernel(x)

        if self.use_reflection:
            # Reflection at the lower boundary
            kernel_lb = stats.gaussian_kde(-values)
            p += kernel_lb.integrate_box_1d(0.0, self.target_volume_pc)
            pdf += kernel_lb(x)

            # Reflection at the upper boundary
            kernel_ub = stats.gaussian_kde(2.0 - values)
            p += kernel_ub.integrate_box_1d(0.0, self.target_volume_pc)
            pdf += kernel_ub(x)

        return p, pdf
GaussianKDEStorageRecorder.register()


class NormalisedGaussianKDEStorageRecorder(NumpyArrayNormalisedStorageRecorder):
    """A recorder that fits a KDE to a normalised time-series of volume.

    This recorder is an extension to `NumpyArrayNormalisedStorageRecorder` which, at the end of a simulation,
    uses kernel density estimation (KDE) to estimate the probability density function of the storage time-series.
    It returns the probability of being at or below zero of the normalised values in the `aggregated_value()` method
    (i.e. used for optimisation). The recorder flattens data from all scenarios before computing the KDE. Therefore,
    a single PDF is produced and returned via `.to_dataframe()`.

    The user can specify an optional resampling (e.g. to create annual minimum time-series) prior to fitting
    the KDE. By default the KDE is reflected at the normalised storage bounds (-1.0 and 1.0) to represent the
    boundedness of the distribution. This can be disabled if required.

    Parameters
    ==========
    resample_freq : str or None
        If not None the resampling frequency used by prior to distribution fitting.
    resample_func : str or None
        If not None the resampling function used prior to distribution fitting.
    num_pdf : int
        Number of points in the PDF estimate. Defaults to 101.
    use_reflection : bool
        Whether to reflect the PDF at the upper and lower normalised bounds (i.e. -1.0 and 1.0 volume) to account for
        the boundedness of the distribution. Defaults to true.
    """
    def __init__(self, *args, **kwargs):
        self.resample_freq = kwargs.pop('resample_freq', None)
        self.resample_func = kwargs.pop('resample_func', None)
        self.use_reflection = kwargs.pop('use_reflection', True)
        self.num_pdf = kwargs.pop('num_pdf', 101)

        super().__init__(*args, **kwargs)
        self._probability_of_target_volume = None
        self._pdf = None

    def reset(self):
        super().reset()
        self._probability_of_target_volume = None
        self._pdf = None

    def finish(self):
        super().finish()

        df = super().to_dataframe()
        # Apply resampling if defined
        if self.resample_func is not None and self.resample_freq is not None:
            df = df.resample(self.resample_freq).agg(self.resample_func)

        x = np.linspace(0.0, 1.0, self.num_pdf)

        # Compute the probability for the target volume
        p, pdf = self._estimate_pdf_and_target_probability(df.values.flatten(), x)

        self._probability_of_target_volume = p
        self._pdf = pandas.DataFrame(data=pdf, index=x)

    def values(self):
        """Return the estimated PDF values."""
        return self._pdf.values

    def to_dataframe(self):
        """ Return a `pandas.DataFrame` of the estimated PDF.
        """
        return self._pdf

    def aggregated_value(self):
        return self._probability_of_target_volume

    def _estimate_pdf_and_target_probability(self, values, x) -> Tuple[float, np.ndarray]:
        """Return a probability of being at below `self.target_volume_pc` and a estimate of the PDF

        This method can (if `self.use_reflection` is truthy) reflect the PDF at the lower and upper boundaries
        to stop the PDF leaking in to infeasible space.
        """
        # Fit a Gaussian KDE
        kernel = stats.gaussian_kde(values)
        p = kernel.integrate_box_1d(-1.0, 0.0)
        pdf = kernel(x)

        if self.use_reflection:
            # Reflection at the lower boundary
            kernel_lb = stats.gaussian_kde(-2.0 - values)
            p += kernel_lb.integrate_box_1d(-1.0, 0.0)
            pdf += kernel_lb(x)

            # Reflection at the upper boundary
            kernel_ub = stats.gaussian_kde(2.0 - values)
            p += kernel_ub.integrate_box_1d(-1.0, 0.0)
            pdf += kernel_ub(x)

        return p, pdf
NormalisedGaussianKDEStorageRecorder.register()



import sys
import pandas
import numpy as np
from functools import wraps
from pywr._core import AbstractNode, AbstractStorage
from ._recorders import *
from ._thresholds import *
from ._hydropower import *
from .events import *
from .calibration import *
from .kde import *
from pywr.h5tools import H5Store
from ..parameter_property import parameter_property
import warnings

class ParameterNameWarning(UserWarning):
    pass

def assert_rec(model, parameter, name=None, get_index=False):
    """Decorator for creating AssertionRecorder objects

    Example
    -------
    @assert_rec(model, parameter)
    def expected_func(timestep, scenario_index):
        return timestep.dayofyear * 2.0
    """
    def assert_rec_(f):
        rec = AssertionRecorder(model, parameter, expected_func=f, name=name, get_index=get_index)
        return f
    return assert_rec_

class AssertionRecorder(Recorder):
    """A recorder that asserts the value of a parameter for testing purposes"""
    def __init__(self, model, parameter, expected_data=None, expected_func=None, get_index=False, **kwargs):
        """
        Parameters
        ----------
        model : pywr.model.Model
        parameter : pywr.parameters.Parameter
        expected_data : np.ndarray[timestep, scenario] (optional)
        expected_func : function
        get_index : bool

        See also
        --------
        pywr.recorders.assert_rec
        """
        super(AssertionRecorder, self).__init__(model, **kwargs)
        self._parameter = None
        self.parameter = parameter
        self.expected_data = expected_data
        self.expected_func = expected_func
        self.get_index = get_index

    parameter = parameter_property("_parameter")

    def setup(self):
        super(AssertionRecorder, self).setup()
        self.count = 0

    def after(self):
        timestep = self.model.timestep
        self.count += 1
        for scenario_index in self.model.scenarios.combinations:
            if self.expected_func:
                expected_value = self.expected_func(timestep, scenario_index)
            elif self.expected_data is not None:
                expected_value = self.expected_data[timestep.index, scenario_index.global_id]
            if self.get_index:
                value = self._parameter.get_index(scenario_index)
            else:
                value = self._parameter.get_value(scenario_index)
            try:
                np.testing.assert_allclose(value, expected_value)
            except AssertionError:
                raise AssertionError("Expected {}, got {} from \"{}\" [timestep={}, scenario={}]".format(expected_value, value, self._parameter.name, timestep.index, scenario_index.global_id))

    def finish(self):
        super(AssertionRecorder, self).finish()
        if sys.exc_info():
            # exception was raised before we had a chance! (e.g. ModelStructureError)
            pass
        elif self.count == 0:
            # this still requires model.run() to have been called...
            raise RuntimeError("AssertionRecorder was never called!")


class CSVRecorder(Recorder):
    """
    A Recorder that saves Node values to a CSV file.

    This class uses the csv package from the Python standard library

    Parameters
    ----------

    model : `pywr.model.Model`
        The model to record nodes from.
    csvfile : str
        The path to the CSV file.
    scenario_index : int
        The scenario index of the model to save.
    nodes : iterable (default=None)
        An iterable of nodes to save data. It defaults to None which is all nodes in the model
    kwargs : Additional keyword arguments to pass to the `csv.writer` object

    """
    def __init__(self, model, csvfile, scenario_index=0, nodes=None, complib=None, complevel=9, **kwargs):
        super(CSVRecorder, self).__init__(model, **kwargs)
        self.csvfile = csvfile
        self.scenario_index = scenario_index
        self.nodes = nodes
        self.csv_kwargs = kwargs.pop('csv_kwargs', {})
        self._node_names = None
        self._fh = None
        self._writer = None
        self.complib = complib
        self.complevel = complevel

    @classmethod
    def load(cls, model, data):
        import os
        url = data.pop("url")
        if not os.path.isabs(url) and model.path is not None:
            url = os.path.join(model.path, url)
        return cls(model, url, **data)

    def setup(self):
        """
        Setup the CSV file recorder.
        """

        if self.nodes is None:
            self._node_names = sorted(self.model.nodes.keys())
        else:
            node_names = []
            for node_ in self.nodes:
                # test if the node name is provided
                if isinstance(node_, str):
                    # lookup node by name
                    node_names.append(node_)
                else:
                    node_names.append((node_.name))
            self._node_names = node_names

    def reset(self):
        import csv
        kwargs = {"newline": "", "encoding": "utf-8"}
        mode = "wt"

        if self.complib == "gzip":
            import gzip
            self._fh = gzip.open(self.csvfile, mode, self.complevel, **kwargs)
        elif self.complib in ("bz2", "bzip2"):
            import bz2
            self._fh = bz2.open(self.csvfile, mode, self.complevel, **kwargs)
        elif self.complib is None:
            self._fh = open(self.csvfile, mode, **kwargs)
        else:
            raise KeyError("Unexpected compression library: {}".format(self.complib))
        self._writer = csv.writer(self._fh, **self.csv_kwargs)
        # Write header data
        row = ["Datetime"] + [name for name in self._node_names]
        self._writer.writerow(row)

    def after(self):
        """
        Write the node values to the CSV file
        """
        values = [self.model.timestepper.current.datetime.isoformat()]
        for node_name in self._node_names:
            node = self.model.nodes[node_name]
            if isinstance(node, AbstractStorage):
                values.append(node.volume[self.scenario_index])
            elif isinstance(node, AbstractNode):
                values.append(node.flow[self.scenario_index])
            else:
                raise ValueError("Unrecognised Node type '{}' for CSV writer".format(type(node)))

        self._writer.writerow(values)

    def finish(self):
        if self._fh:
            self._fh.close()
CSVRecorder.register()


class TablesRecorder(Recorder):
    """
    A recorder that saves to PyTables CArray

    This Recorder creates a CArray for every node passed to the constructor.
    Each CArray stores the data for all scenarios on the specific node. This
    is useful for analysis of Node statistics across multiple scenarios.
    """
    def __init__(self, model, h5file, nodes=None, parameters=None, where='/', time='/time',
                 routes_flows=None, routes='/routes', scenarios='/scenarios', **kwargs):
        """

        Parameters
        ----------
        model : `pywr.model.Model`
            The model to record nodes from.
        h5file : tables.File or filename
            The tables file handle or filename to attach the CArray objects to. If a
            filename is given the object will open and close the file handles.
        nodes : iterable or None
            Nodes to save in the tables database. Can be an iterable of Node objects or
            node names. It can also be a iterable of tuples with a node specific where
            keyword as the first item and a Node object or name as the second item. If
            an iterable of tuples is provided then the node specific where keyword is
            used in preference to the where keyword (see below).
        parameters : iterable or None
            Parameters to save. Similar to the nodes keyword, except refers to Parameter
            objects or names thereof.
        where : string
            Default path to create the CArrays inside the database.
        time : string
            Default full node path to save a time tables.Table. If None no table is created.
        scenarios : string
            Default full node path to save a scenarios tables.Table. If None no table is created.
        routes_flows : string
            Relative (to `where`) node path to save the routes flow CArray. If None (default) no array is created.
        routes : string
            Full node path to save the routes tables.Table. If None not table is created.
        filter_kwds : dict
            Filter keywords to pass to tables.open_file when opening a file.
        mode : string
            Model argument to pass to tables.open_file. Defaults to 'w'
        metadata : dict
            Dict of user defined attributes to save on the root node (`root._v_attrs`)
        create_directories : bool
            If a file path is given and create_directories is True then attempt to make the intermediate
            directories. This uses os.makedirs() underneath.
        """
        self.filter_kwds = kwargs.pop('filter_kwds', {})
        self.mode = kwargs.pop('mode', 'w')
        self.metadata = kwargs.pop('metadata', {})
        self.create_directories = kwargs.pop('create_directories', False)

        title = kwargs.pop('title', None)
        if title is None:
            try:
                title = model.metadata['title']
            except KeyError:
                title = ''
        self.title = title
        super(TablesRecorder, self).__init__(model, **kwargs)

        self.h5file = h5file
        self.h5store = None
        self._arrays = {}
        self.nodes = nodes
        self.where = where
        self.time = time
        self.scenarios = scenarios
        self.routes = routes
        self.routes_flows = routes_flows

        # Enable saving routes in the solver.
        if routes_flows:
            self.model.solver.save_routes_flows = True

        self.parameters = []
        if parameters:
            for parameter in parameters:
                self._add_parameter(parameter)

        self._arrays = None
        self._time_table = None
        self._routes_flow_array = None

    def _add_parameter(self, parameter):
        try:
            where, param = parameter
        except (TypeError, ValueError):
            where = None
            param = parameter
        if isinstance(param, str):
            from ..parameters import load_parameter
            param = load_parameter(self.model, param)
        if not param.name:
            raise ValueError("Can only record named Parameter objects")
        if where is None:
            name = param.name.replace("/", "_")
            if name != param.name:
                warnings.warn(
                    "Recorded parameter has \"/\" in name, replaced with \"_\" to avoid creation of subgroup: {}".format(param.name),
                    ParameterNameWarning
                )
            where = self.where + "/" + name
        where = where.replace("//", "/")
        self.children.add(param)
        self.parameters.append((where, param))

    def _remove_parameter(self, parameter):
        if isinstance(parameter, str):
            parameter = self.model.parameters[parameter]
        index = None
        for n, (where, param) in enumerate(self.parameters):
            if param is parameter:
                index = n
        if index is None:
            raise KeyError("Parameter is not in TablesRecorder: {}".format(parameter))
        self.parameters.pop(index)
        self.children.remove(parameter)

    @classmethod
    def load(cls, model, data):
        import os
        url = data.pop("url")
        if not os.path.isabs(url) and model.path is not None:
            url = os.path.join(model.path, url)
        return cls(model, url, **data)

    def setup(self):
        """
        Setup the tables
        """
        from pywr.parameters import IndexParameter
        import tables

        # The first dimension is the number of timesteps.
        # The following dimensions are sized per scenario
        scenario_shape = list(self.model.scenarios.shape)
        shape = [len(self.model.timestepper)] + scenario_shape

        self.h5store = H5Store(self.h5file, self.filter_kwds, self.mode, title=self.title, metadata=self.metadata,
                               create_directories=self.create_directories)

        # Create a CArray for each node
        self._arrays = {}

        # Default to all nodes if None given.
        if self.nodes is None:
            nodes = [((self.where + "/" + n.name).replace("//", "/"), n) for n in self.model.nodes.values()]
        else:
            nodes = []
            for n in self.nodes:

                try:
                    where, node = n
                except (TypeError, ValueError):
                    node = n
                    where = self.where + "/" + node

                # Accept a str, and lookup node by name instead.
                if isinstance(node, str):
                    node = self.model.nodes[node]
                # Otherwise assume it is a node object anyway

                where = where.replace("//", "/")
                nodes.append((where, node))

        if self.parameters is not None:
            nodes.extend(self.parameters)

        self._nodes = nodes

        for where, node in self._nodes:
            if isinstance(node, IndexParameter):
                atom = tables.Int32Atom()
            else:
                atom = tables.Float64Atom()
            group_name, node_name = where.rsplit("/", 1)
            if group_name == "":
                group_name = "/"
            self.h5store.file.create_carray(group_name, node_name, atom, shape, createparents=True)

        # Create scenario tables
        if self.scenarios is not None:
            group_name, node_name = self.scenarios.rsplit('/', 1)
            if group_name == "":
                group_name = "/"
            description = {
                # TODO make string length configurable
                'name': tables.StringCol(1024),
                'size': tables.Int64Col()
            }
            tbl = self.h5store.file.create_table(group_name, node_name, description=description, createparents=True)
            # Now add the scenarios
            entry = tbl.row
            for scenario in self.model.scenarios.scenarios:
                entry['name'] = scenario.name.encode('utf-8')
                entry['size'] = scenario.size
                entry.append()
            tbl.flush()

            if self.model.scenarios.user_combinations is not None:
                description = {s.name: tables.Int64Col() for s in self.model.scenarios.scenarios}
                tbl = self.h5store.file.create_table(group_name, 'scenario_combinations', description=description)
                entry = tbl.row
                for comb in self.model.scenarios.user_combinations:
                    for s, i in zip(self.model.scenarios.scenarios, comb):
                        entry[s.name] = i
                    entry.append()
                tbl.flush()

        self.h5store = None

    def reset(self):
        import tables

        mode = "r+"  # always need to append, as file already created in setup
        self.h5store = H5Store(self.h5file, self.filter_kwds, mode)
        self._arrays = {}
        for where, node in self._nodes:
            self._arrays[node] = self.h5store.file.get_node(where)

        self._time_table = None
        # Create time table
        # This is created in reset so that the table is always recreated
        if self.time is not None:
            group_name, node_name = self.time.rsplit('/', 1)
            if group_name == "":
                group_name = "/"
            description = {c: tables.Int64Col() for c in ('year', 'month', 'day', 'index')}

            try:
                self.h5store.file.remove_node(group_name, node_name)
            except tables.NoSuchNodeError:
                pass
            finally:
                self._time_table = self.h5store.file.create_table(group_name, node_name,
                                                                  description=description,
                                                                  createparents=True)

        self._routes_flow_array = None
        if self.routes_flows is not None:
            # Create a CArray for the flows
            # The first dimension is the number of timesteps.
            # The second dimension is the number of routes
            # The following dimensions are sized per scenario
            scenario_shape = list(self.model.scenarios.shape)
            shape = [len(self.model.timestepper), len(self.model.solver.routes)] + scenario_shape
            atom = tables.Float64Atom()

            try:
                self.h5store.file.remove_node(self.where, self.routes_flows)
            except tables.NoSuchNodeError:
                pass
            finally:
                self._routes_flow_array = self.h5store.file.create_carray(self.where, self.routes_flows, atom, shape, createparents=True)

            # Create routes table. This must be done in reset
            if self.routes is not None:
                group_name, node_name = self.routes.rsplit('/', 1)
                if group_name == "":
                    group_name = "/"

                description = {
                    # TODO make string length configurable
                    'start': tables.StringCol(1024),
                    'end': tables.StringCol(1024),
                }
                try:
                    self.h5store.file.remove_node(group_name, node_name)
                except tables.NoSuchNodeError:
                    pass
                finally:
                    tbl = self.h5store.file.create_table(group_name, node_name, description=description, createparents=True)

                entry = tbl.row
                for route in self.model.solver.routes:
                    node_first = route[0]
                    node_last = route[-1]

                    if node_first.parent is not None:
                        node_first = node_first.parent
                    if node_last.parent is not None:
                        node_last = node_last.parent

                    entry['start'] = node_first.name.encode('utf-8')
                    entry['end'] = node_last.name.encode('utf-8')
                    entry.append()

                tbl.flush()

    def after(self):
        """
        Save data to the tables
        """
        from pywr._core import AbstractNode, AbstractStorage
        from pywr.parameters import Parameter, IndexParameter
        scenario_shape = list(self.model.scenarios.shape)
        ts = self.model.timestepper.current
        idx = ts.index

        if self._time_table is not None:
            entry = self._time_table.row
            entry['year'] = ts.year
            entry['month'] = ts.month
            entry['day'] = ts.day
            entry['index'] = idx
            entry.append()

        for node, ca in self._arrays.items():
            if isinstance(node, AbstractStorage):
                ca[idx, :] = np.reshape(node.volume, scenario_shape)
            elif isinstance(node, AbstractNode):
                ca[idx, :] = np.reshape(node.flow, scenario_shape)
            elif isinstance(node, IndexParameter):
                a = node.get_all_indices()
                ca[idx, :] = np.reshape(a, scenario_shape)
            elif isinstance(node, Parameter):
                a = node.get_all_values()
                ca[idx, :] = np.reshape(a, scenario_shape)
            else:
                raise ValueError("Unrecognised Node type '{}' for TablesRecorder".format(type(node)))

        if self._routes_flow_array is not None:
            routes_shape = [len(self.model.solver.routes), ] + scenario_shape
            self._routes_flow_array[idx, ...] = np.reshape(self.model.solver.routes_flows_array.T, routes_shape)

    def finish(self):
        if self._time_table is not None:
            self._time_table.flush()
        self.h5store = None
        self._arrays = {}
        self._routes_flow_array = None

    @staticmethod
    def generate_dataframes(h5file, time='/time', scenarios='/scenarios'):
        """Helper function to generate pandas dataframes from `TablesRecorder` data.

        Parameters
        h5file : str
            A path to a H5 file created by `TablesRecorder`.
        time : str
            The internal table that contains the time information (default "/time")
        scenarios : str
            The internal table that contains the scenario information (default "/scenarios")
        """
        store = H5Store(h5file, mode='r')

        # Get the time information
        if time:
            time_table = store.file.get_node(time)
            index = pandas.to_datetime({k: time_table.col(k) for k in ('year', 'month', 'day')})
        else:
            index = None

        # Get the scenario information
        if scenarios:
            scenarios_table = store.file.get_node(scenarios)
            scenarios = pandas.DataFrame({k: scenarios_table.col(k) for k in ('name', 'size')})
            columns = pandas.MultiIndex.from_product(
                [range(row['size']) for _, row in scenarios.iterrows()],
                names=[row['name'].decode() for _, row in scenarios.iterrows()]
            )
        else:
            columns = None

        for node in store.file.walk_nodes('/', 'CArray'):
            data = node.read()
            data = data.reshape((data.shape[0], -1))
            df = pandas.DataFrame(data, index=index, columns=columns)
            yield node._v_name, df

TablesRecorder.register()



from ..core import Model
import uuid
import logging
logger = logging.getLogger(__name__)


def cache_variable_parameters(model):
    variables = []
    variable_map = [0, ]
    for var in model.variables:
        size = var.double_size + var.integer_size

        if size <= 0:
            raise ValueError('Variable parameter "{}" does not have a size > 0.'.format(var.name))

        variable_map.append(variable_map[-1] + size)
        variables.append(var)

    return variables, variable_map


def cache_constraints(model):
    constraints = []
    for r in model.constraints:
        constraints.append(r)

    return constraints


def cache_objectives(model):
    # This is done to make sure the order is fixed during optimisation.
    objectives = []
    for r in model.objectives:
        objectives.append(r)
    return objectives


# Global variables for individual processes to cache the model and some of its data.
# The cache is keyed by a UID for each `BaseOptimisationWrapper`
class ModelCache:
    def __init__(self):
        self.model = None
        self.variables = None
        self.variable_map = None
        self.objectives = None
        self.constraints = None
MODEL_CACHE = {}


class BaseOptimisationWrapper(object):
    """ A helper class for running pywr optimisations with platypus.
    """
    def __init__(self, pywr_model_json, *args, **kwargs):
        uid = kwargs.pop('uid', None)
        self.pywr_model_klass = kwargs.pop("model_klass", Model)
        super(BaseOptimisationWrapper, self).__init__(*args, **kwargs)
        self.pywr_model_json = pywr_model_json

        if uid is None:
            uid = uuid.uuid4().hex  # Create a unique ID for caching.
        self.uid = uid
        self.run_stats = None

    # The following properties enable attribute caching when repeat execution of the same model is undertaken.
    @property
    def _cached(self):
        global MODEL_CACHE
        try:
            cache = MODEL_CACHE[self.uid]
        except KeyError:
            model = self.make_model()
            model.setup()

            cache = ModelCache()
            cache.model = model
            cache.variables, cache.variable_map = cache_variable_parameters(model)
            cache.objectives = cache_objectives(model)
            cache.constraints = cache_constraints(model)
            MODEL_CACHE[self.uid] = cache
        return cache

    @property
    def model(self):
        return self._cached.model

    @property
    def model_variables(self):
        return self._cached.variables

    @property
    def model_variable_map(self):
        return self._cached.variable_map

    @property
    def model_objectives(self):
        return self._cached.objectives

    @property
    def model_constraints(self):
        return self._cached.constraints

    def make_model(self):
        m = self.pywr_model_klass.load(self.pywr_model_json)
        # Apply any user defined changes to the model
        self.customise_model(m)
        return m

    def customise_model(self, model):
        pass  # By default there is no customisation.


def clear_global_model_cache():
    """ Clear the module level model cache. """
    global MODEL_CACHE
    MODEL_CACHE = {}



from .._core cimport Timestep, Scenario, ScenarioIndex, AbstractNode, Storage, AbstractStorage
from ._parameters cimport Parameter, IndexParameter


cdef class Polynomial1DParameter(Parameter):
    cdef public double[:] coefficients
    cdef AbstractNode _other_node
    cdef AbstractStorage _storage_node
    cdef public bint use_proportional_volume
    cdef Parameter _parameter
    cdef double offset
    cdef double scale

cdef class Polynomial2DStorageParameter(Parameter):
    cdef public double[:, :] coefficients
    cdef AbstractStorage _storage_node
    cdef public bint use_proportional_volume
    cdef Parameter _parameter
    # Scaling parameters
    cdef double storage_offset
    cdef double storage_scale
    cdef double parameter_offset
    cdef double parameter_scale

